-------------------------------------------------------------------------------
--
-- Copyright 2023
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------
--
-- Author: E. Kooistra
-- Purpose:
-- . SDP beamlet data output (BDO) destinations register
-- Description:
--              _________
--             |mm_fields|
--             |         |
-- reg_copi -->|  slv_out|------- _wr ----
-- reg_cipo <--|         |               |
--             |         |               v
--             |   slv_in|<--+--- _rd <--+<---- _ro
--             |_________|   |
--                           -----------------> output value
--
-- where:
-- . _ro = actual nof_destinations, actual nof_blocks_per_packet
-- . output value = sdp_bdo_destinations
--
-- References:
-- 1 https://plm.astron.nl/polarion/#/project/LOFAR2System/wiki/L2%20Interface%20Control%20Documents/SC%20to%20SDP%20ICD
-- 2 https://support.astron.nl/confluence/display/L2M/L4+SDPFW+Decision%3A+Multiple+beamlet+output+destinations
--

library IEEE, common_lib, mm_lib;
  use IEEE.std_logic_1164.all;
  use common_lib.common_pkg.all;
  use common_lib.common_mem_pkg.all;
  use common_lib.common_field_pkg.all;
  use work.sdp_pkg.all;
  use work.sdp_bdo_pkg.all;

entity sdp_bdo_destinations_reg is
  port (
    -- Clocks and reset
    mm_rst   : in  std_logic;
    mm_clk   : in  std_logic;

    dp_clk   : in  std_logic;
    dp_rst   : in  std_logic;

    reg_copi : in  t_mem_copi;
    reg_cipo : out t_mem_cipo;

    -- sdp info
    sdp_bdo_destinations : out t_sdp_bdo_destinations_info
  );
end sdp_bdo_destinations_reg;

architecture str of sdp_bdo_destinations_reg is
  constant c_field_arr : t_common_field_arr(c_sdp_bdo_destinations_info_nof_hdr_fields - 1 downto 0) :=
    ( (field_name_pad("nof_blocks_per_packet_act"), "RO",  8, field_default(c_sdp_cep_nof_blocks_per_packet)),
      (field_name_pad("nof_destinations_act"),      "RO",  8, field_default(1)),
      (field_name_pad("nof_destinations"),          "RW",  8, field_default(1)),

      (field_name_pad("udp_destination_port_15"),   "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_14"),   "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_13"),   "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_12"),   "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_11"),   "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_10"),   "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_9"),    "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_8"),    "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_7"),    "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_6"),    "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_5"),    "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_4"),    "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_3"),    "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_2"),    "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_1"),    "RW", 16, field_default(0)),
      (field_name_pad("udp_destination_port_0"),    "RW", 16, field_default(0)),

      (field_name_pad("ip_destination_address_15"), "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_14"), "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_13"), "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_12"), "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_11"), "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_10"), "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_9"),  "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_8"),  "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_7"),  "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_6"),  "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_5"),  "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_4"),  "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_3"),  "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_2"),  "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_1"),  "RW", 32, field_default(0)),
      (field_name_pad("ip_destination_address_0"),  "RW", 32, field_default(0)),

      (field_name_pad("eth_destination_mac_15"),    "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_14"),    "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_13"),    "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_12"),    "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_11"),    "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_10"),    "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_9"),     "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_8"),     "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_7"),     "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_6"),     "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_5"),     "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_4"),     "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_3"),     "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_2"),     "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_1"),     "RW", 48, field_default(0)),
      (field_name_pad("eth_destination_mac_0"),     "RW", 48, field_default(0)) );

  signal mm_fields_in  : std_logic_vector(field_slv_in_len(c_field_arr) - 1 downto 0);
  signal mm_fields_out : std_logic_vector(field_slv_out_len(c_field_arr) - 1 downto 0);

  signal sdp_bdo_destinations_rd : t_sdp_bdo_destinations_info;
  signal sdp_bdo_destinations_wr : t_sdp_bdo_destinations_info;

  signal nof_destinations_act      : natural := 1;
  signal nof_blocks_per_packet_act : natural := c_sdp_cep_nof_blocks_per_packet;
begin
  sdp_bdo_destinations <= sdp_bdo_destinations_rd;

  p_sdp_bdo_destinations_rd : process(sdp_bdo_destinations_wr,
                                      nof_destinations_act,
                                      nof_blocks_per_packet_act)
  begin
    -- default write assign all fields
    sdp_bdo_destinations_rd <= sdp_bdo_destinations_wr;

    -- overrule the read only fields
    sdp_bdo_destinations_rd.nof_destinations_act      <= nof_destinations_act;
    sdp_bdo_destinations_rd.nof_blocks_per_packet_act <= nof_blocks_per_packet_act;
  end process;

  u_mm_fields: entity mm_lib.mm_fields
  generic map(
    g_use_slv_in_val  => false,  -- use FALSE to save logic when always slv_in_val='1'
    g_field_arr       => c_field_arr
  )
  port map (
    mm_clk     => mm_clk,
    mm_rst     => mm_rst,

    mm_mosi    => reg_copi,
    mm_miso    => reg_cipo,

    slv_clk    => dp_clk,
    slv_rst    => dp_rst,

    slv_in     => mm_fields_in,
    slv_in_val => '1',

    slv_out    => mm_fields_out
  );

  -- add "RO" fields to mm_fields
  mm_fields_in(field_hi(c_field_arr, "nof_destinations_act") downto field_lo(c_field_arr, "nof_destinations_act")) <= to_uvec(sdp_bdo_destinations_rd.nof_destinations_act, 8);
  mm_fields_in(field_hi(c_field_arr, "nof_blocks_per_packet_act") downto field_lo(c_field_arr, "nof_blocks_per_packet_act")) <= to_uvec(sdp_bdo_destinations_rd.nof_blocks_per_packet_act, 8);

  -- get "RW" fields from mm_fields
  sdp_bdo_destinations_wr.eth_destination_mac_arr(0)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_0") downto field_lo(c_field_arr, "eth_destination_mac_0"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(1)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_1") downto field_lo(c_field_arr, "eth_destination_mac_1"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(2)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_2") downto field_lo(c_field_arr, "eth_destination_mac_2"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(3)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_3") downto field_lo(c_field_arr, "eth_destination_mac_3"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(4)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_4") downto field_lo(c_field_arr, "eth_destination_mac_4"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(5)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_5") downto field_lo(c_field_arr, "eth_destination_mac_5"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(6)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_6") downto field_lo(c_field_arr, "eth_destination_mac_6"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(7)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_7") downto field_lo(c_field_arr, "eth_destination_mac_7"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(8)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_8") downto field_lo(c_field_arr, "eth_destination_mac_8"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(9)  <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_9") downto field_lo(c_field_arr, "eth_destination_mac_9"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(10) <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_10") downto field_lo(c_field_arr, "eth_destination_mac_10"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(11) <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_11") downto field_lo(c_field_arr, "eth_destination_mac_11"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(12) <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_12") downto field_lo(c_field_arr, "eth_destination_mac_12"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(13) <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_13") downto field_lo(c_field_arr, "eth_destination_mac_13"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(14) <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_14") downto field_lo(c_field_arr, "eth_destination_mac_14"));
  sdp_bdo_destinations_wr.eth_destination_mac_arr(15) <= mm_fields_out(field_hi(c_field_arr, "eth_destination_mac_15") downto field_lo(c_field_arr, "eth_destination_mac_15"));

  sdp_bdo_destinations_wr.ip_destination_address_arr(0)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_0") downto field_lo(c_field_arr, "ip_destination_address_0"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(1)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_1") downto field_lo(c_field_arr, "ip_destination_address_1"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(2)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_2") downto field_lo(c_field_arr, "ip_destination_address_2"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(3)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_3") downto field_lo(c_field_arr, "ip_destination_address_3"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(4)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_4") downto field_lo(c_field_arr, "ip_destination_address_4"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(5)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_5") downto field_lo(c_field_arr, "ip_destination_address_5"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(6)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_6") downto field_lo(c_field_arr, "ip_destination_address_6"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(7)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_7") downto field_lo(c_field_arr, "ip_destination_address_7"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(8)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_8") downto field_lo(c_field_arr, "ip_destination_address_8"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(9)  <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_9") downto field_lo(c_field_arr, "ip_destination_address_9"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(10) <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_10") downto field_lo(c_field_arr, "ip_destination_address_10"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(11) <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_11") downto field_lo(c_field_arr, "ip_destination_address_11"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(12) <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_12") downto field_lo(c_field_arr, "ip_destination_address_12"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(13) <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_13") downto field_lo(c_field_arr, "ip_destination_address_13"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(14) <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_14") downto field_lo(c_field_arr, "ip_destination_address_14"));
  sdp_bdo_destinations_wr.ip_destination_address_arr(15) <= mm_fields_out(field_hi(c_field_arr, "ip_destination_address_15") downto field_lo(c_field_arr, "ip_destination_address_15"));

  sdp_bdo_destinations_wr.udp_destination_port_arr(0)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_0") downto field_lo(c_field_arr, "udp_destination_port_0"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(1)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_1") downto field_lo(c_field_arr, "udp_destination_port_1"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(2)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_2") downto field_lo(c_field_arr, "udp_destination_port_2"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(3)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_3") downto field_lo(c_field_arr, "udp_destination_port_3"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(4)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_4") downto field_lo(c_field_arr, "udp_destination_port_4"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(5)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_5") downto field_lo(c_field_arr, "udp_destination_port_5"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(6)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_6") downto field_lo(c_field_arr, "udp_destination_port_6"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(7)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_7") downto field_lo(c_field_arr, "udp_destination_port_7"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(8)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_8") downto field_lo(c_field_arr, "udp_destination_port_8"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(9)  <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_9") downto field_lo(c_field_arr, "udp_destination_port_9"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(10) <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_10") downto field_lo(c_field_arr, "udp_destination_port_10"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(11) <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_11") downto field_lo(c_field_arr, "udp_destination_port_11"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(12) <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_12") downto field_lo(c_field_arr, "udp_destination_port_12"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(13) <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_13") downto field_lo(c_field_arr, "udp_destination_port_13"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(14) <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_14") downto field_lo(c_field_arr, "udp_destination_port_14"));
  sdp_bdo_destinations_wr.udp_destination_port_arr(15) <= mm_fields_out(field_hi(c_field_arr, "udp_destination_port_15") downto field_lo(c_field_arr, "udp_destination_port_15"));

  sdp_bdo_destinations_wr.nof_destinations <= to_uint(mm_fields_out(field_hi(c_field_arr, "nof_destinations") downto field_lo(c_field_arr, "nof_destinations")));

  -- Register the read only actual values, to ease timing closure
  p_dp_clk : process(dp_clk)
    constant c_look_up_table : t_natural_arr(1 to c_sdp_bdo_nof_destinations_max) :=
      func_sdp_bdo_nof_blocks_per_packet_look_up_table;
  begin
    if rising_edge(dp_clk) then
      nof_destinations_act <= func_sdp_bdo_parse_nof_destinations(sdp_bdo_destinations_wr.nof_destinations);
      nof_blocks_per_packet_act <= c_look_up_table(nof_destinations_act);
    end if;
  end process;
end str;
