-------------------------------------------------------------------------------
--
-- Copyright 2023
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: E. Kooistra
-- Purpose:
-- . This package contains sdp beamlet data output (BDO) specific constants.
-- Description: See [1]
-- References:
-- . [1] https://support.astron.nl/confluence/display/L2M/L4+SDPFW+Decision%3A+Multiple+beamlet+output+destinations
-------------------------------------------------------------------------------
library IEEE, common_lib;
use IEEE.std_logic_1164.all;
use common_lib.common_pkg.all;
use work.sdp_pkg.all;

package sdp_bdo_pkg is
  -- Beamlet data output (BDO) for multiple destinations
  constant c_sdp_bdo_nof_destinations_max   : natural := 16;
  constant c_sdp_bdo_reorder_nof_blocks_max : natural := largest(16, c_sdp_cep_nof_blocks_per_packet);

  constant c_sdp_bdo_destinations_info_nof_hdr_fields : natural := c_sdp_bdo_nof_destinations_max * 3 + 3;  -- = 52 fields

  type t_sdp_bdo_destinations_info is record
    eth_destination_mac_arr     : t_slv_48_arr(c_sdp_bdo_nof_destinations_max - 1 downto 0);
    ip_destination_address_arr  : t_slv_32_arr(c_sdp_bdo_nof_destinations_max - 1 downto 0);
    udp_destination_port_arr    : t_slv_16_arr(c_sdp_bdo_nof_destinations_max - 1 downto 0);
    nof_destinations            : natural;
    nof_destinations_act        : natural;
    nof_blocks_per_packet_act   : natural;
  end record;

  constant t_sdp_bdo_destinations_info_rst : t_sdp_bdo_destinations_info :=
    ( (others => (others => '0')),
      (others => (others => '0')),
      (others => (others => '0')), 0, 0, 0 );

  -- Determine actual nof_destinations and actual nof_blocks_per_packet
  function func_sdp_bdo_parse_nof_destinations(nof_destinations : natural) return natural;
  function func_sdp_bdo_nof_blocks_per_packet_look_up_table return t_natural_arr;

  -- Determine nof_beamlets_per_block per destination
  function func_sdp_sdo_nof_beamlets_per_block_first_destination_look_up_table return t_natural_arr;
  function func_sdp_sdo_nof_beamlets_per_block_last_destination_look_up_table return t_natural_arr;
  function func_sdp_sdo_beamlet_index_per_destination_look_up_matrix return t_natural_matrix;
end package sdp_bdo_pkg;

package body sdp_bdo_pkg is
  function func_sdp_bdo_parse_nof_destinations(nof_destinations : natural) return natural is
  begin
    -- Parse input nof_destinations value
    if nof_destinations = 0 then
      return 1;
    elsif nof_destinations > c_sdp_bdo_nof_destinations_max then
      return c_sdp_bdo_nof_destinations_max;
    else
      return nof_destinations;
    end if;
  end func_sdp_bdo_parse_nof_destinations;

  function func_sdp_bdo_nof_blocks_per_packet_look_up_table return t_natural_arr is
    variable v_arr : t_natural_arr(1 to c_sdp_bdo_nof_destinations_max);
  begin
    -- Determine nof_blocks_per_packet as function of number of destinations D. Use
    -- look up table to precalculate the values as constants.
    -- . Use same number of nof_blocks_per_packet for each destination.
    -- . In total there are maximum c_sdp_bdo_reorder_nof_blocks_max = 16 blocks to
    --   distribute over D destinations, so floor(16 / D) per destination yields:
    --     D = 1:16 --> 16, 8, 5, 4, 3, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1
    for D in 1 to c_sdp_bdo_nof_destinations_max loop
      v_arr(D) := c_sdp_bdo_reorder_nof_blocks_max / D;
    end loop;
    -- . With 1 destination c_sdp_cep_nof_blocks_per_packet = 4 can fit in a jumbo frame.
    -- . With D destinations D * c_sdp_cep_nof_blocks_per_packet can fit in a jumbo frame,
    --   because the number of beamlets per destination reduces by D.
    -- . Taking smallest yields the actual maximum number of blocks per packet
    --   per destination, as function of number of destinations D:
    --     D = 1:16 --> 4, 8, 5, 4, 3, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1
    for D in 1 to c_sdp_bdo_nof_destinations_max loop
      v_arr(D) := smallest(v_arr(D), D * c_sdp_cep_nof_blocks_per_packet);
    end loop;
    return v_arr;
  end func_sdp_bdo_nof_blocks_per_packet_look_up_table;

  function func_sdp_sdo_nof_beamlets_per_block_first_destination_look_up_table return t_natural_arr is
    variable v_first_arr : t_natural_arr(1 to c_sdp_bdo_nof_destinations_max);
  begin
    -- Determine nof_beamlets_per_block for the first 1:DN-1 destinations, as
    -- function of number of destinations DN. Use look up table to precalculate
    -- the values as constants.
    -- . In total there are c_sdp_S_sub_bf = 488 dual polarization beamlets to
    --   distribute over DN destinations, so ceil(488 / DN) yields the number of
    --   blocks for the first 1:DN-1 destinations:
    --     DN = 1:16 --> v_first_arr = 488, 244, 163, 122, 98, 82, 70, 61, 55, 49, 45, 41, 38, 35, 33, 31
    for DN in 1 to c_sdp_bdo_nof_destinations_max loop
      v_first_arr(DN) := ceil_div(c_sdp_S_sub_bf, DN);
    end loop;
    return v_first_arr;
  end func_sdp_sdo_nof_beamlets_per_block_first_destination_look_up_table;

  function func_sdp_sdo_nof_beamlets_per_block_last_destination_look_up_table return t_natural_arr is
    variable v_first_arr : t_natural_arr(1 to c_sdp_bdo_nof_destinations_max) :=
                             func_sdp_sdo_nof_beamlets_per_block_first_destination_look_up_table;
    variable v_last_arr  : t_natural_arr(1 to c_sdp_bdo_nof_destinations_max);
  begin
    -- Determine nof_beamlets_per_block for the last destination with index DN,
    -- as function of number of destinations DN. Use look up table to
    -- precalculate the values as constants.
    -- . In total there are c_sdp_S_sub_bf = 488 dual polarization beamlets to
    --   distribute over DN destinations, so 488 - (DN-1) * ceil(488 / DN)
    --   beamlets remain for the last destination:
    --     DN = 1:16 --> v_first_arr = 488, 244, 163, 122, 98, 82, 70, 61, 55, 49, 45, 41, 38, 35, 33, 31
    --     DN = 1:16 --> v_last_arr  = 488, 244, 162, 122, 96, 78, 68, 61, 48, 47, 38, 37, 32, 33, 26, 23
    --
    -- Remark:
    -- . The v_last_arr may be < v_first_arr - 1, so the last destination may
    --   contain much less beamlets than the others. In combination with
    --   dp_packet_unmerge it is not feasible to distribute the beamlets evenly
    --   over all destinations, using v_hi beamlets for some first destinations
    --   and v_lo = v_hi - 1 for the remaining destinations. This is because
    --   dp_packet_unmerge can only unmerge the same packet length for N - 1
    --   blocks and then unmerge the remaining data in the last block until the
    --   eop.
    --
    for DN in 1 to c_sdp_bdo_nof_destinations_max loop
      v_last_arr(DN) := c_sdp_S_sub_bf - (DN - 1) * v_first_arr(DN);
    end loop;
    return v_last_arr;
  end func_sdp_sdo_nof_beamlets_per_block_last_destination_look_up_table;

  function func_sdp_sdo_beamlet_index_per_destination_look_up_matrix return t_natural_matrix is
    constant c_len_arr   : t_natural_arr(1 to c_sdp_bdo_nof_destinations_max) :=
                             func_sdp_sdo_nof_beamlets_per_block_first_destination_look_up_table;
    variable v_index_mat : t_natural_matrix(1 to c_sdp_bdo_nof_destinations_max,
                                            0 to c_sdp_bdo_nof_destinations_max - 1);
    variable v_beamlet_index : natural;
    variable v_step          : natural;
  begin
    -- Determine beamlet index of first beamlet in packet per destination with
    -- index DN, as function of number of destinations DN. Use look up table to
    -- precalculate the values as constants.
    -- . Beamlet index for first destination starts at 0
    -- . Beamlet index for the other destinations increments with number of
    --   beamlets per previous destination given by c_len_arr.
    --
    -- * rows: nof_destinations DN
    -- * columns: destination index DI
    --
    -- DI:   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,  13,  14,  15
    -- DN:
    --  1    0,   0.....................................................................0
    --  2    0, 244,   0                                                                .
    --  3    0, 163, 326,   0                                                           .
    --  4    0, 122, 244, 366,   0                                                      .
    --  5    0,  98, 196, 294, 392,   0                                                 .
    --  6    0,  82, 164, 246, 328, 410,   0                                            .
    --  7    0,  70, 140, 210, 280, 350, 420,   0                                       .
    --  8    0,  61, 122, 183, 244, 305, 366, 427,   0                                  .
    --  9    0,  55, 110, 165, 220, 275, 330, 385, 440,   0                             .
    -- 10    0,  49,  98, 147, 196, 245, 294, 343, 392, 441,   0                        .
    -- 11    0,  45,  90, 135, 180, 225, 270, 315, 360, 405, 450,   0                   .
    -- 12    0,  41,  82, 123, 164, 205, 246, 287, 328, 369, 410, 451,   0              .
    -- 13    0,  38,  76, 114, 152, 190, 228, 266, 304, 342, 380, 418, 456,   0         .
    -- 14    0,  35,  70, 105, 140, 175, 210, 245, 280, 315, 350, 385, 420, 455,   0    .
    -- 15    0,  33,  66,  99, 132, 165, 198, 231, 264, 297, 330, 363, 396, 429, 462,   0
    -- 16    0,  31,  62,  93, 124, 155, 186, 217, 248, 279, 310, 341, 372, 403, 434, 465
    --
    -- Equivalent Python code to produce matrix:
    --
    --   c_len_arr = [488, 244, 163, 122, 98, 82, 70, 61, 55, 49, 45, 41, 38, 35, 33, 31]
    --   for DN in range(16):
    --       lineStr = '%2d ' % (DN + 1)
    --       v_beamlet_index = 0
    --       v_step = c_len_arr[DN]
    --       for DI in range(16):
    --           if v_beamlet_index < 488:
    --               lineStr += '%4d,' % v_beamlet_index
    --           v_beamlet_index += v_step
    --       print(lineStr)
    --
    for DN in 1 to c_sdp_bdo_nof_destinations_max loop
      v_beamlet_index := 0;
      v_step := c_len_arr(DN);
      for DI in 0 to c_sdp_bdo_nof_destinations_max - 1 loop
        if v_beamlet_index < c_sdp_S_sub_bf then
          v_index_mat(DN, DI) := v_beamlet_index;
        end if;
        v_beamlet_index := v_beamlet_index + v_step;
      end loop;
    end loop;
    return v_index_mat;
  end func_sdp_sdo_beamlet_index_per_destination_look_up_matrix;
end sdp_bdo_pkg;
